
#ifndef __MDL_LIMITS_H_
#define __MDL_LIMITS_H_


namespace mdl
{


// Maximum number of LODs per model
const int MAX_LODS = 8;


// Maximum number of bones per vertex
const int MAX_BONES_PER_VERTEX = 3;


}


#endif

