#include <osgDB/Registry>

USE_SERIALIZER_WRAPPER(osgAnimation_Action)
USE_SERIALIZER_WRAPPER(osgAnimation_ActionAnimation)
USE_SERIALIZER_WRAPPER(osgAnimation_ActionBlendIn)
USE_SERIALIZER_WRAPPER(osgAnimation_ActionBlendOut)
USE_SERIALIZER_WRAPPER(osgAnimation_ActionStripAnimation)
USE_SERIALIZER_WRAPPER(osgAnimation_Animation)
USE_SERIALIZER_WRAPPER(osgAnimation_AnimationManagerBase)
USE_SERIALIZER_WRAPPER(osgAnimation_BasicAnimationManager)
USE_SERIALIZER_WRAPPER(osgAnimation_Bone)
USE_SERIALIZER_WRAPPER(osgAnimation_MorphGeometry)
USE_SERIALIZER_WRAPPER(osgAnimation_RigComputeBoundingBoxCallback)
USE_SERIALIZER_WRAPPER(osgAnimation_RigGeometry)
USE_SERIALIZER_WRAPPER(osgAnimation_RigTransform)
USE_SERIALIZER_WRAPPER(osgAnimation_Skeleton)
USE_SERIALIZER_WRAPPER(osgAnimation_StackedMatrixElement)
USE_SERIALIZER_WRAPPER(osgAnimation_StackedQuaternionElement)
USE_SERIALIZER_WRAPPER(osgAnimation_StackedRotateAxisElement)
USE_SERIALIZER_WRAPPER(osgAnimation_StackedScaleElement)
USE_SERIALIZER_WRAPPER(osgAnimation_StackedTransformElement)
USE_SERIALIZER_WRAPPER(osgAnimation_StackedTranslateElement)
USE_SERIALIZER_WRAPPER(osgAnimation_Timeline)
USE_SERIALIZER_WRAPPER(osgAnimation_TimelineAnimationManager)
USE_SERIALIZER_WRAPPER(osgAnimation_UpdateBone)
USE_SERIALIZER_WRAPPER(osgAnimation_UpdateMaterial)
USE_SERIALIZER_WRAPPER(osgAnimation_UpdateMatrixTransform)
USE_SERIALIZER_WRAPPER(osgAnimation_UpdateMorph)
USE_SERIALIZER_WRAPPER(osgAnimation_UpdateSkeleton)
USE_SERIALIZER_WRAPPER(osgAnimation_UpdateMorphGeometry)
USE_SERIALIZER_WRAPPER(osgAnimation_UpdateRigGeometry)
USE_SERIALIZER_WRAPPER(osgAnimation_UpdateFloatUniform)
USE_SERIALIZER_WRAPPER(osgAnimation_UpdateMatrixfUniform)
USE_SERIALIZER_WRAPPER(osgAnimation_UpdateVec2fUniform)
USE_SERIALIZER_WRAPPER(osgAnimation_UpdateVec3fUniform)
USE_SERIALIZER_WRAPPER(osgAnimation_UpdateVec4fUniform)


extern "C" void wrapper_serializer_library_osgAnimation(void) {}

