#include <osgDB/Registry>

USE_SERIALIZER_WRAPPER(osgVolume_AlphaFuncProperty)
USE_SERIALIZER_WRAPPER(osgVolume_CompositeLayer)
USE_SERIALIZER_WRAPPER(osgVolume_CompositeProperty)
USE_SERIALIZER_WRAPPER(osgVolume_FixedFunctionTechnique)
USE_SERIALIZER_WRAPPER(osgVolume_ImageDetails)
USE_SERIALIZER_WRAPPER(osgVolume_ImageLayer)
USE_SERIALIZER_WRAPPER(osgVolume_IsoSurfaceProperty)
USE_SERIALIZER_WRAPPER(osgVolume_Layer)
USE_SERIALIZER_WRAPPER(osgVolume_LightingProperty)
USE_SERIALIZER_WRAPPER(osgVolume_Locator)
USE_SERIALIZER_WRAPPER(osgVolume_MaximumIntensityProjectionProperty)
USE_SERIALIZER_WRAPPER(osgVolume_Property)
USE_SERIALIZER_WRAPPER(osgVolume_PropertyAdjustmentCallback)
USE_SERIALIZER_WRAPPER(osgVolume_RayTracedTechnique)
USE_SERIALIZER_WRAPPER(osgVolume_SampleDensityProperty)
USE_SERIALIZER_WRAPPER(osgVolume_ScalarProperty)
USE_SERIALIZER_WRAPPER(osgVolume_SwitchProperty)
USE_SERIALIZER_WRAPPER(osgVolume_TransferFunctionProperty)
USE_SERIALIZER_WRAPPER(osgVolume_TransparencyProperty)
USE_SERIALIZER_WRAPPER(osgVolume_Volume)
USE_SERIALIZER_WRAPPER(osgVolume_VolumeTechnique)
USE_SERIALIZER_WRAPPER(osgVolume_VolumeTile)

extern "C" void wrapper_serializer_library_osgVolume(void) {}

