#include <osgDB/Registry>

USE_DOTOSGWRAPPER(CompositeProperty_Proxy)
USE_DOTOSGWRAPPER(FixedFunctionTechnique_Proxy)
USE_DOTOSGWRAPPER(ImageLayer_Proxy)
USE_DOTOSGWRAPPER(Layer_Proxy)
USE_DOTOSGWRAPPER(Locator_Proxy)
USE_DOTOSGWRAPPER(Property_Proxy)
USE_DOTOSGWRAPPER(MaximumImageProjectionProperty_Proxy)
USE_DOTOSGWRAPPER(LightingProperty_Proxy)
USE_DOTOSGWRAPPER(PropertyAdjustmentCallback_Proxy)
USE_DOTOSGWRAPPER(RayTracedTechnique_Proxy)
USE_DOTOSGWRAPPER(ScalarProperty_Proxy)
USE_DOTOSGWRAPPER(IsoSurfaceProperty_Proxy)
USE_DOTOSGWRAPPER(AlphaFuncProperty_Proxy)
USE_DOTOSGWRAPPER(SampleDensityProperty_Proxy)
USE_DOTOSGWRAPPER(TransparencyProperty_Proxy)
USE_DOTOSGWRAPPER(SwitchProperty_Proxy)
USE_DOTOSGWRAPPER(TransferFunctionProperty_Proxy)
USE_DOTOSGWRAPPER(Volume_Proxy)
USE_DOTOSGWRAPPER(VolumeTile_Proxy)

extern "C" void dotosgwrapper_library_osgVolume(void) {}

