#include <osgDB/Registry>

USE_SERIALIZER_WRAPPER(osgManipulator_AntiSquish)
USE_SERIALIZER_WRAPPER(osgManipulator_CompositeDragger)
USE_SERIALIZER_WRAPPER(osgManipulator_Dragger)
USE_SERIALIZER_WRAPPER(osgManipulator_RotateCylinderDragger)
USE_SERIALIZER_WRAPPER(osgManipulator_RotateSphereDragger)
USE_SERIALIZER_WRAPPER(osgManipulator_Scale1DDragger)
USE_SERIALIZER_WRAPPER(osgManipulator_Scale2DDragger)
USE_SERIALIZER_WRAPPER(osgManipulator_ScaleAxisDragger)
USE_SERIALIZER_WRAPPER(osgManipulator_TabBoxDragger)
USE_SERIALIZER_WRAPPER(osgManipulator_TabBoxTrackballDragger)
USE_SERIALIZER_WRAPPER(osgManipulator_TabPlaneDragger)
USE_SERIALIZER_WRAPPER(osgManipulator_TabPlaneTrackballDragger)
USE_SERIALIZER_WRAPPER(osgManipulator_TrackballDragger)
USE_SERIALIZER_WRAPPER(osgManipulator_Translate1DDragger)
USE_SERIALIZER_WRAPPER(osgManipulator_Translate2DDragger)
USE_SERIALIZER_WRAPPER(osgManipulator_TranslateAxisDragger)
USE_SERIALIZER_WRAPPER(osgManipulator_TranslatePlaneDragger)

extern "C" void wrapper_serializer_library_osgManipulator(void) {}

