#include <osgDB/Registry>

USE_DOTOSGWRAPPER(BlinkSequence_Proxy)
USE_DOTOSGWRAPPER(BlinkSequence_SequenceGroup_Proxy)
USE_DOTOSGWRAPPER(g_DOFTransformProxy)
USE_DOTOSGWRAPPER(g_ImpostorProxy)
USE_DOTOSGWRAPPER(LightPointNode_Proxy)
USE_DOTOSGWRAPPER(g_simSwitchProxy)
USE_DOTOSGWRAPPER(ObjectRecordData_Proxy)
USE_DOTOSGWRAPPER(g_OverlayNodeProxy)
USE_DOTOSGWRAPPER(AzimSector_Proxy)
USE_DOTOSGWRAPPER(ElevationSector_Proxy)
USE_DOTOSGWRAPPER(AzimElevationSector_Proxy)
USE_DOTOSGWRAPPER(ConeSector_Proxy)
USE_DOTOSGWRAPPER(DirectionalSector_Proxy)
USE_DOTOSGWRAPPER(ShapeAttributeList_Proxy)
USE_DOTOSGWRAPPER(g_VisibilityGroupProxy)

extern "C" void dotosgwrapper_library_osgSim(void) {}

