#include <osgShadow/ShadowTechnique>
#include <osgDB/ObjectWrapper>
#include <osgDB/InputStream>
#include <osgDB/OutputStream>

REGISTER_OBJECT_WRAPPER( osgShadow_ShadowTechnique,
                         0,
                         osgShadow::ShadowTechnique,
                         "osg::Object osgShadow::ShadowTechnique" )
{
}
