#line 2 "ConfigLexer.cpp"
/* A lexical scanner generated by flex */

/* Scanner skeleton version:
 * $Header: /cvs/Producer/Producer/src/ConfigLexer.cpp,v 1.23 2006/09/06 22:56:30 don Exp $
 */

#define FLEX_SCANNER
#define YY_FLEX_MAJOR_VERSION 2
#define YY_FLEX_MINOR_VERSION 5

#include <osg/Config>

#if defined(_MSC_VER) && defined(OSG_DISABLE_MSVC_WARNINGS)
    // disable '=' : conversion from 'int' to 'char', possible loss of data style warnings
    #pragma warning( disable : 4244)
#endif


/* cfront 1.2 defines "c_plusplus" instead of "__cplusplus" */
#ifdef c_plusplus
#ifndef __cplusplus
#define __cplusplus
#endif
#endif


#ifdef __cplusplus

#include <stdlib.h>
#include <iostream>
#ifndef WIN32
#include <unistd.h>
#endif

/* Use prototypes in function declarations. */
#define YY_USE_PROTOS

/* The "const" storage-class-modifier is valid. */
#define YY_USE_CONST

#else	/* ! __cplusplus */

#if __STDC__

#define YY_USE_PROTOS
#define YY_USE_CONST

#endif	/* __STDC__ */
#endif	/* ! __cplusplus */

#ifdef __TURBOC__
 #pragma warn -rch
 #pragma warn -use
#include <io.h>
#include <stdlib.h>
#define YY_USE_CONST
#define YY_USE_PROTOS
#endif

#ifdef YY_USE_CONST
#define yyconst const
#else
#define yyconst
#endif


#ifdef YY_USE_PROTOS
#define YY_PROTO(proto) proto
#else
#define YY_PROTO(proto) ()
#endif

/* Returned upon end-of-file. */
#define YY_NULL 0

/* Promotes a possibly negative, possibly signed char to an unsigned
 * integer for use as an array index.  If the signed char is negative,
 * we want to instead treat it as an 8-bit unsigned char, hence the
 * double cast.
 */
#define YY_SC_TO_UI(c) ((unsigned int) (unsigned char) c)

/* Enter a start condition.  This macro really ought to take a parameter,
 * but we do it the disgusting crufty way forced on us by the ()-less
 * definition of BEGIN.
 */
#define BEGIN yy_start = 1 + 2 *

/* Translate the current start state into a value that can be later handed
 * to BEGIN to return to the state.  The YYSTATE alias is for lex
 * compatibility.
 */
#define YY_START ((yy_start - 1) / 2)
#define YYSTATE YY_START

/* Action number for EOF rule of a given start state. */
#define YY_STATE_EOF(state) (YY_END_OF_BUFFER + state + 1)

/* Special action meaning "start processing a new file". */
#define YY_NEW_FILE yyrestart( yyin )

#define YY_END_OF_BUFFER_CHAR 0

/* Size of default input buffer. */
#define YY_BUF_SIZE 16384

typedef struct yy_buffer_state *YY_BUFFER_STATE;

extern int yyleng;

#define EOB_ACT_CONTINUE_SCAN 0
#define EOB_ACT_END_OF_FILE 1
#define EOB_ACT_LAST_MATCH 2

/* The funky do-while in the following #define is used to turn the definition
 * int a single C statement (which needs a semi-colon terminator).  This
 * avoids problems with code like:
 *
 * 	if ( condition_holds )
 *		yyless( 5 );
 *	else
 *		do_something_else();
 *
 * Prior to using the do-while the compiler would get upset at the
 * "else" because it interpreted the "if" statement as being all
 * done when it reached the ';' after the yyless() call.
 */

/* Return all but the first 'n' matched characters back to the input stream. */

#define yyless(n) \
	do \
		{ \
		/* Undo effects of setting up yytext. */ \
		*yy_cp = yy_hold_char; \
		YY_RESTORE_YY_MORE_OFFSET \
		yy_c_buf_p = yy_cp = yy_bp + n - YY_MORE_ADJ; \
		YY_DO_BEFORE_ACTION; /* set up yytext again */ \
		} \
	while ( 0 )

#define unput(c) yyunput( c, yytext_ptr )

/* The following is because we cannot portably get our hands on size_t
 * (without autoconf's help, which isn't available because we want
 * flex-generated scanners to compile on their own).
 */
typedef unsigned int yy_size_t;


struct yy_buffer_state
	{
	std::istream* yy_input_file;

	char *yy_ch_buf;		/* input buffer */
	char *yy_buf_pos;		/* current position in input buffer */

	/* Size of input buffer in bytes, not including room for EOB
	 * characters.
	 */
	yy_size_t yy_buf_size;

	/* Number of characters read into yy_ch_buf, not including EOB
	 * characters.
	 */
	int yy_n_chars;

	/* Whether we "own" the buffer - i.e., we know we created it,
	 * and can realloc() it to grow it, and should free() it to
	 * delete it.
	 */
	int yy_is_our_buffer;

	/* Whether this is an "interactive" input source; if so, and
	 * if we're using stdio for input, then we want to use getc()
	 * instead of fread(), to make sure we stop fetching input after
	 * each newline.
	 */
	int yy_is_interactive;

	/* Whether we're considered to be at the beginning of a line.
	 * If so, '^' rules will be active on the next match, otherwise
	 * not.
	 */
	int yy_at_bol;

	/* Whether to try to fill the input buffer when we reach the
	 * end of it.
	 */
	int yy_fill_buffer;

	int yy_buffer_status;
#define YY_BUFFER_NEW 0
#define YY_BUFFER_NORMAL 1
	/* When an EOF's been seen but there's still some text to process
	 * then we mark the buffer as YY_EOF_PENDING, to indicate that we
	 * shouldn't try reading from the input source any more.  We might
	 * still have a bunch of tokens to match, though, because of
	 * possible backing-up.
	 *
	 * When we actually see the EOF, we change the status to "new"
	 * (via yyrestart()), so that the user can continue scanning by
	 * just pointing yyin at a new input file.
	 */
#define YY_BUFFER_EOF_PENDING 2
	};


/* We provide macros for accessing buffer states in case in the
 * future we want to put the buffer states in a more general
 * "scanner state".
 */
#define YY_CURRENT_BUFFER yy_current_buffer



static void *yy_flex_alloc YY_PROTO(( yy_size_t ));
static void *yy_flex_realloc YY_PROTO(( void *, yy_size_t ));
static void yy_flex_free YY_PROTO(( void * ));

#define yy_new_buffer yy_create_buffer

#define yy_set_interactive(is_interactive) \
	{ \
	if ( ! yy_current_buffer ) \
		yy_current_buffer = yy_create_buffer( yyin, YY_BUF_SIZE ); \
	yy_current_buffer->yy_is_interactive = is_interactive; \
	}

#define yy_set_bol(at_bol) \
	{ \
	if ( ! yy_current_buffer ) \
		yy_current_buffer = yy_create_buffer( yyin, YY_BUF_SIZE ); \
	yy_current_buffer->yy_at_bol = at_bol; \
	}

#define YY_AT_BOL() (yy_current_buffer->yy_at_bol)


#define yywrap() 1
#define YY_SKIP_YYWRAP
typedef unsigned char YY_CHAR;
#define yytext_ptr yytext
#define YY_INTERACTIVE

#include "FlexLexer.h"


/* Done after the current pattern has been matched and before the
 * corresponding action - sets up yytext.
 */
#define YY_DO_BEFORE_ACTION \
	yytext_ptr = yy_bp; \
	yyleng = (int) (yy_cp - yy_bp); \
	yy_hold_char = *yy_cp; \
	*yy_cp = '\0'; \
	yy_c_buf_p = yy_cp;

#define YY_NUM_RULES 83
#define YY_END_OF_BUFFER 84
static yyconst short int yy_accept[588] =
    {   0,
        0,    0,   84,   83,    3,    2,    1,    6,    7,   83,
       79,   83,   78,   78,    8,    9,   83,   83,   83,   83,
       83,   83,   83,   83,   83,   83,   83,   83,   83,   83,
       83,   83,   83,   83,   83,   83,   83,   83,    4,    5,
       79,   78,   78,   79,   79,    1,    0,   80,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
       75,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,   76,    0,    0,    0,

        0,    0,   81,   77,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,   82,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,   52,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,   14,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,

        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,   13,    0,    0,   57,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,   62,    0,    0,    0,   63,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,   37,
       49,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,   64,   59,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,   60,    0,    0,   16,   36,
        0,    0,    0,    0,    0,    0,    0,   30,    0,    0,

        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,   35,    0,   55,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,   46,    0,
        0,    0,   28,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,   34,    0,    0,    0,    0,    0,    0,    0,    0,
       18,   45,   47,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,   11,    0,    0,    0,
        0,    0,    0,    0,    0,   20,    0,    0,    0,    0,

        0,    0,    0,    0,   67,   39,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,   10,   53,   54,
        0,    0,    0,    0,   61,    0,    0,    0,    0,    0,
        0,   21,    0,    0,    0,   58,    0,   22,    0,    0,
       19,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,   33,
        0,    0,    0,    0,   17,   12,   48,    0,    0,    0,
        0,    0,    0,   56,    0,   65,    0,    0,    0,    0,
        0,    0,    0,    0,   71,    0,    0,   42,    0,    0,
        0,    0,    0,    0,   15,   40,    0,    0,   41,   66,

        0,   43,    0,    0,    0,   23,    0,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,   31,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
       24,    0,   38,    0,   51,    0,   29,   72,    0,    0,
        0,    0,    0,   26,    0,    0,    0,    0,    0,    0,
       73,    0,   32,   27,   25,    0,   70,    0,    0,    0,
        0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
        0,    0,   50,   44,    0,    0,    0,   68,    0,    0,
        0,    0,   74,    0,    0,   69,    0
    } ;

static yyconst int yy_ec[256] =
    {   0,
        1,    1,    1,    1,    1,    1,    1,    1,    2,    3,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    2,    4,    5,    6,    1,    1,    1,    1,    1,
        1,    1,    7,    8,    9,   10,   11,   12,   13,   13,
       13,   13,   13,   13,   13,   13,   13,   14,   15,    1,
        1,    1,    1,    1,   16,   17,   18,   19,   20,   21,
       22,   23,   24,    1,    1,   25,   26,   27,   28,   29,
        1,   30,   31,   32,   33,   34,   35,   36,   37,   38,
        1,    1,    1,    1,   39,    1,   40,   41,   42,   43,

       44,   45,   46,   47,   48,   49,    1,   50,   51,   52,
       53,   54,    1,   55,   56,   57,   58,   59,   60,   61,
       62,    1,   63,    1,   64,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,

        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1
    } ;

static yyconst int yy_meta[65] =
    {   0,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    2,    2,    1,    1,    2,    2,    2,    2,    2,
        2,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    2,
        2,    2,    2,    2,    2,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1
    } ;

static yyconst short int yy_base[589] =
    {   0,
        0,    0,  667,  668,  668,  668,  668,  668,  668,   55,
       57,  655,   61,   67,  668,  668,   58,   60,   46,   69,
      610,  634,  610,  610,   46,  617,   71,   55,   87,  102,
       48,  612,   77,  619,  605,   69,  602,  612,  668,  668,
      115,  135,  141,  145,  668,  668,  126,  668,    0,  637,
      625,  617,  619,  630,  595,  598,  604,  591,  617,  612,
      588,  603,  584,  621,  584,  585,  604,  585,  579,  614,
      668,  589,  576,  588,  598,  575,  573,   31,  609,  610,
      594,   80,  568,  598,  603,   60,  565,   97,  569,  576,
      564,  578,  561,  589,  563,  564,  668,  568,  554,  555,

      160,  130,  151,    0,  577,  586,  569,  587,  585,  562,
      560,  563,  545,  569,  583,  545,  538,  541,  576,  538,
      536,  573,  536,  544,  534,  542,  533,  566,  530,  528,
      558,  534,  543,  565,  541,  536,  535,  537,  547,  141,
      525,  530,  542,  517,  531,  524,  514,  524,  515,  508,
      546,  521,  507,  518,  162,  164,  535,  544,  542,  519,
      537,  512,  500,  499,  500,  529,  526,  500,  509,  491,
      520,  494,  488,  519,  668,  490,  498,  488,  485,  518,
      484,  511,  478,  491,  503,  668,  151,  514,  488,  474,
      505,  511,  508,  483,  482,  477,  480,  468,  472,  477,

      480,  463,  478,  489,  463,  471,  475,  474,  479,  480,
      480,  454,  468,  489,  455,  466,  484,  463,  461,  443,
      461,  459,  154,  668,  455,  440,  668,  448,  475,  450,
      435,  442,  449,  466,  461,  466,  432,  431,  441,  464,
      459,  454,  668,  429,  429,  157,  668,  435,  425,  434,
      426,  425,  439,  413,  171,  441,  450,  446,  430,  668,
      446,  414,  445,  434,  447,  401,  412,  410,  429,  408,
      403,  413,  668,  668,  413,  425,  412,  403,  395,  394,
      412,  422,  431,  407,  163,  668,  161,  421,  668,  668,
      391,  400,  395,  410,  410,  173,  400,  415,  399,  407,

      411,  410,  404,  413,  408,  410,  392,  398,  373,  377,
      368,  401,  391,  668,  379,  668,  398,  377,  376,  377,
      374,  378,  359,  358,  366,  365,  392,  391,  394,  349,
      350,  354,  668,  389,  366,  354,  351,  358,  354,  338,
      346,  354,  340,  377,  363,  350,  364,  359,  371,  371,
      351,  368,  367,  362,  332,  331,  333,  344,  360,  348,
      341,  668,  338,  320,  346,  343,  326,  325,  318,  318,
      668,  668,  668,  330,  314,  336,  334,  335,  321,  308,
      303,  307,  305,  317,  304,  314,  668,  320,  314,  332,
      334,  333,  313,  331,  320,  668,  311,  290,  292,  296,

      325,  323,  281,  322,  668,  301,  296,  302,  279,  283,
      286,  283,  293,  288,  288,  310,  306,  668,  668,  668,
      285,  271,  283,  179,  668,  297,  268,  308,  284,  295,
      290,  668,  289,  299,  264,  668,  286,  668,  296,  261,
      668,  262,  270,  283,  267,  260,  247,  278,  257,  266,
      244,  283,  265,  262,  257,  250,  259,  254,  277,  256,
      256,  263,  254,  268,  668,  668,  668,  249,  260,  245,
      242,  239,  266,  668,  223,  668,  240,  239,  240,  224,
      260,  259,  235,  226,  668,  225,  223,  668,  222,  242,
      248,  240,  232,  214,  668,  668,  218,  212,  668,  668,

      224,  668,  221,  206,  233,  668,  218,  243,  216,  216,
      212,  233,  218,  231,  234,  209,  208,  207,  193,  668,
      194,  217,  204,  193,  190,  200,  193,  204,  221,  202,
      668,  195,  668,  196,  197,  192,  668,  668,  184,  194,
      178,  188,  211,  668,  210,  177,  171,  175,  200,  174,
      668,  193,  668,  668,  668,  193,  668,  176,  168,  180,
      161,  174,  167,  173,  163,  159,  171,  168,  167,  167,
      163,  150,  668,  668,  150,  165,  164,  668,  161,  146,
      139,  106,  668,   79,   79,  668,  668,   92
    } ;

static yyconst short int yy_def[589] =
    {   0,
      587,    1,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  588,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,

      587,  587,  587,  588,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,

      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,

      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,

      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,

      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,    0,  587
    } ;

static yyconst short int yy_nxt[733] =
    {   0,
        4,    5,    6,    7,    8,    7,    4,    9,   10,   11,
       12,   13,   14,   15,   16,   17,   18,   19,   20,    4,
       21,   22,   23,   24,   25,   26,    4,   27,   28,   29,
       30,   31,    4,   32,   33,    4,    4,    4,    4,    4,
        4,    4,    4,    4,   34,    4,    4,    4,    4,    4,
        4,   35,   36,    4,    4,    4,   37,    4,    4,    4,
        4,   38,   39,   40,   41,   67,   42,   43,   44,   44,
       41,   75,   43,   43,  131,   50,   41,   45,   43,   43,
       47,   48,   51,  132,   53,   56,   47,   48,   59,   68,
       52,   70,   54,  104,   91,   57,   60,   71,   76,  141,

       94,   45,   92,   58,   47,   48,   79,   77,   80,   78,
       47,   48,   55,   98,  142,   72,   61,   84,   81,  136,
       71,   49,  586,   62,   95,   73,   44,   44,  585,   74,
       82,  137,  102,   85,  102,   45,  144,  103,  103,   83,
      145,  103,  103,   86,   41,   87,   43,   43,   88,   89,
       41,  584,   43,   43,   47,   48,   44,   44,   90,   45,
       47,   48,  103,  103,  101,   45,  155,  192,  155,  271,
      193,  156,  156,  156,  156,  156,  156,  235,   47,   48,
      236,  292,  583,  272,   47,   48,  301,  302,  101,   45,
      293,  333,  303,  331,  332,   49,  457,  582,  341,  334,

      304,  342,  581,  580,  579,  578,  577,  576,  458,  575,
      574,  573,  572,  571,  570,  569,  568,  567,  566,  565,
      564,  563,  562,  561,  560,  559,  558,  557,  556,  555,
      554,  553,  552,  551,  550,  549,  548,  547,  546,  545,
      544,  543,  542,  541,  540,  539,  538,  537,  536,  535,
      534,  533,  532,  531,  530,  529,  528,  527,  526,  525,
      524,  523,  522,  521,  520,  519,  518,  517,  516,  515,
      514,  513,  512,  511,  510,  509,  508,  507,  506,  505,
      504,  503,  502,  501,  500,  499,  498,  497,  496,  495,
      494,  493,  492,  491,  490,  489,  488,  487,  486,  485,

      484,  483,  482,  481,  480,  479,  478,  477,  476,  475,
      474,  473,  472,  471,  470,  469,  468,  467,  466,  465,
      464,  463,  462,  461,  460,  459,  456,  455,  454,  453,
      452,  451,  450,  449,  448,  447,  446,  445,  444,  443,
      442,  441,  440,  439,  438,  437,  436,  435,  434,  433,
      432,  431,  430,  429,  428,  427,  426,  425,  424,  423,
      422,  421,  420,  419,  418,  417,  416,  415,  414,  413,
      412,  411,  410,  409,  408,  407,  406,  405,  404,  403,
      402,  401,  400,  399,  398,  397,  396,  395,  394,  393,
      392,  391,  390,  389,  388,  387,  386,  385,  384,  383,

      382,  381,  380,  379,  378,  377,  376,  375,  374,  373,
      372,  371,  370,  369,  368,  367,  366,  365,  364,  363,
      362,  361,  360,  359,  358,  357,  356,  355,  354,  353,
      352,  351,  350,  349,  348,  347,  346,  345,  344,  343,
      340,  339,  338,  337,  336,  335,  330,  329,  328,  327,
      326,  325,  324,  323,  322,  321,  320,  319,  318,  317,
      316,  315,  314,  313,  312,  311,  310,  309,  308,  307,
      306,  305,  300,  299,  298,  297,  296,  295,  294,  291,
      290,  289,  288,  287,  286,  285,  284,  283,  282,  281,
      280,  279,  278,  277,  276,  275,  274,  273,  270,  269,

      268,  267,  266,  265,  264,  263,  262,  261,  260,  259,
      258,  257,  256,  255,   97,  254,  253,  252,  251,  250,
      249,  248,  247,  246,  245,  244,  243,  242,  241,  240,
      239,  238,  237,  234,  233,  232,  231,  230,  229,  228,
      227,  226,  225,  224,  223,  222,  221,  220,  219,  218,
      217,  216,  215,  214,  213,  212,  211,  210,  209,  208,
      207,   71,  206,  205,  204,  203,  202,  201,  200,  199,
      198,  197,  196,  195,  194,  191,  190,  189,  188,  187,
      186,  185,  184,  183,  182,  181,  180,  179,  178,  177,
      176,  175,  174,  173,  172,  171,  170,  169,  168,  167,

      166,  165,  164,  163,  162,  161,  160,  159,  158,  157,
       71,  154,   97,  153,  152,  151,  150,  149,  148,  147,
      146,  143,  140,  139,  138,  135,  134,  133,  130,  129,
      128,  127,  126,  125,   97,  124,  123,  122,  121,  120,
      119,  118,  117,  116,  115,  114,  113,  112,  111,  110,
      109,  108,  107,  106,  105,  100,   99,   97,   96,   93,
       69,   66,   65,   64,   63,   46,  587,    3,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,

      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587
    } ;

static yyconst short int yy_chk[733] =
    {   0,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
        1,    1,    1,    1,   10,   25,   10,   10,   11,   11,
       13,   28,   13,   13,   78,   17,   14,   11,   14,   14,
       13,   13,   17,   78,   18,   19,   14,   14,   20,   25,
       17,   27,   18,  588,   31,   19,   20,   27,   28,   86,

       33,   11,   31,   19,   13,   13,   29,   28,   29,   28,
       14,   14,   18,   36,   86,   27,   20,   30,   29,   82,
       36,   13,  585,   20,   33,   27,   41,   41,  584,   27,
       29,   82,   47,   30,   47,   41,   88,   47,   47,   29,
       88,  102,  102,   30,   42,   30,   42,   42,   30,   30,
       43,  582,   43,   43,   42,   42,   44,   44,   30,   41,
       43,   43,  103,  103,   44,   44,  101,  140,  101,  223,
      140,  101,  101,  155,  155,  156,  156,  187,   42,   42,
      187,  246,  581,  223,   43,   43,  255,  255,   44,   44,
      246,  287,  255,  285,  285,   42,  424,  580,  296,  287,

      255,  296,  579,  577,  576,  575,  572,  571,  424,  570,
      569,  568,  567,  566,  565,  564,  563,  562,  561,  560,
      559,  558,  556,  552,  550,  549,  548,  547,  546,  545,
      543,  542,  541,  540,  539,  536,  535,  534,  532,  530,
      529,  528,  527,  526,  525,  524,  523,  522,  521,  519,
      518,  517,  516,  515,  514,  513,  512,  511,  510,  509,
      508,  507,  505,  504,  503,  501,  498,  497,  494,  493,
      492,  491,  490,  489,  487,  486,  484,  483,  482,  481,
      480,  479,  478,  477,  475,  473,  472,  471,  470,  469,
      468,  464,  463,  462,  461,  460,  459,  458,  457,  456,

      455,  454,  453,  452,  451,  450,  449,  448,  447,  446,
      445,  444,  443,  442,  440,  439,  437,  435,  434,  433,
      431,  430,  429,  428,  427,  426,  423,  422,  421,  417,
      416,  415,  414,  413,  412,  411,  410,  409,  408,  407,
      406,  404,  403,  402,  401,  400,  399,  398,  397,  395,
      394,  393,  392,  391,  390,  389,  388,  386,  385,  384,
      383,  382,  381,  380,  379,  378,  377,  376,  375,  374,
      370,  369,  368,  367,  366,  365,  364,  363,  361,  360,
      359,  358,  357,  356,  355,  354,  353,  352,  351,  350,
      349,  348,  347,  346,  345,  344,  343,  342,  341,  340,

      339,  338,  337,  336,  335,  334,  332,  331,  330,  329,
      328,  327,  326,  325,  324,  323,  322,  321,  320,  319,
      318,  317,  315,  313,  312,  311,  310,  309,  308,  307,
      306,  305,  304,  303,  302,  301,  300,  299,  298,  297,
      295,  294,  293,  292,  291,  288,  284,  283,  282,  281,
      280,  279,  278,  277,  276,  275,  272,  271,  270,  269,
      268,  267,  266,  265,  264,  263,  262,  261,  259,  258,
      257,  256,  254,  253,  252,  251,  250,  249,  248,  245,
      244,  242,  241,  240,  239,  238,  237,  236,  235,  234,
      233,  232,  231,  230,  229,  228,  226,  225,  222,  221,

      220,  219,  218,  217,  216,  215,  214,  213,  212,  211,
      210,  209,  208,  207,  206,  205,  204,  203,  202,  201,
      200,  199,  198,  197,  196,  195,  194,  193,  192,  191,
      190,  189,  188,  185,  184,  183,  182,  181,  180,  179,
      178,  177,  176,  174,  173,  172,  171,  170,  169,  168,
      167,  166,  165,  164,  163,  162,  161,  160,  159,  158,
      157,  154,  153,  152,  151,  150,  149,  148,  147,  146,
      145,  144,  143,  142,  141,  139,  138,  137,  136,  135,
      134,  133,  132,  131,  130,  129,  128,  127,  126,  125,
      124,  123,  122,  121,  120,  119,  118,  117,  116,  115,

      114,  113,  112,  111,  110,  109,  108,  107,  106,  105,
      100,   99,   98,   96,   95,   94,   93,   92,   91,   90,
       89,   87,   85,   84,   83,   81,   80,   79,   77,   76,
       75,   74,   73,   72,   70,   69,   68,   67,   66,   65,
       64,   63,   62,   61,   60,   59,   58,   57,   56,   55,
       54,   53,   52,   51,   50,   38,   37,   35,   34,   32,
       26,   24,   23,   22,   21,   12,    3,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,

      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587,  587,  587,  587,  587,  587,  587,  587,  587,
      587,  587
    } ;

/* The intent behind this definition is that it'll catch
 * any uses of REJECT which flex missed.
 */
#define REJECT reject_used_but_not_detected
#define yymore() yymore_used_but_not_detected
#define YY_MORE_ADJ 0
#define YY_RESTORE_YY_MORE_OFFSET
#line 1 ".././ConfigLexer.l"
#define INITIAL 0
/* -*-c++-*- Producer - Copyright (C) 2001-2004  Don Burns
 *
 * This library is open source and may be redistributed and/or modified under
 * the terms of the OpenSceneGraph Public License (OSGPL) version 0.0 or
 * (at your option) any later version.  The full license is in LICENSE file
 * included with this distribution, and on the openscenegraph.org website.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * OpenSceneGraph Public License for more details.
 */
#line 15 ".././ConfigLexer.l"
#include <stdio.h>
#include <ctype.h>
#include "ConfigParser.h"

//#define DEBUG
#ifdef DEBUG
#define REPORT printf(" %s\n", yytext);
#else
#define REPORT
#endif
/*%option yyclass="ConfigParser"*/
#line 705 "ConfigLexer.cpp"

/* Macros after this point can all be overridden by user definitions in
 * section 1.
 */

#ifndef YY_SKIP_YYWRAP
#ifdef __cplusplus
extern "C" int yywrap YY_PROTO(( void ));
#else
extern int yywrap YY_PROTO(( void ));
#endif
#endif


#ifndef yytext_ptr
static void yy_flex_strncpy YY_PROTO(( char *, yyconst char *, int ));
#endif

#ifdef YY_NEED_STRLEN
static int yy_flex_strlen YY_PROTO(( yyconst char * ));
#endif

#ifndef YY_NO_INPUT
#endif

#if YY_STACK_USED
static int yy_start_stack_ptr = 0;
static int yy_start_stack_depth = 0;
static int *yy_start_stack = 0;
#ifndef YY_NO_PUSH_STATE
static void yy_push_state YY_PROTO(( int new_state ));
#endif
#ifndef YY_NO_POP_STATE
static void yy_pop_state YY_PROTO(( void ));
#endif
#ifndef YY_NO_TOP_STATE
static int yy_top_state YY_PROTO(( void ));
#endif

#else
#define YY_NO_PUSH_STATE 1
#define YY_NO_POP_STATE 1
#define YY_NO_TOP_STATE 1
#endif

#ifdef YY_MALLOC_DECL
YY_MALLOC_DECL
#else
#if __STDC__
#ifndef __cplusplus
#include <stdlib.h>
#endif
#else
/* Just try to get by without declaring the routines.  This will fail
 * miserably on non-ANSI systems for which sizeof(size_t) != sizeof(int)
 * or sizeof(void*) != sizeof(int).
 */
#endif
#endif

/* Amount of stuff to slurp up with each read. */
#ifndef YY_READ_BUF_SIZE
#define YY_READ_BUF_SIZE 8192
#endif

/* Copy whatever the last rule matched to the standard output. */

#ifndef ECHO
#define ECHO LexerOutput( yytext, yyleng )
#endif

/* Gets input and stuffs it into "buf".  number of characters read, or YY_NULL,
 * is returned in "result".
 */
#ifndef YY_INPUT
#define YY_INPUT(buf,result,max_size) \
	if ( (result = LexerInput( (char *) buf, max_size )) < 0 ) \
		YY_FATAL_ERROR( "input in flex scanner failed" );
#endif

/* No semi-colon after return; correct usage is to write "yyterminate();" -
 * we don't want an extra ';' after the "return" because that will cause
 * some compilers to complain about unreachable statements.
 */
#ifndef yyterminate
#define yyterminate() return YY_NULL
#endif

/* Number of entries by which start-condition stack grows. */
#ifndef YY_START_STACK_INCR
#define YY_START_STACK_INCR 25
#endif

/* Report a fatal error. */
#ifndef YY_FATAL_ERROR
#define YY_FATAL_ERROR(msg) LexerError( msg )
#endif

/* Default declaration of generated scanner - a define so the user can
 * easily add parameters.
 */
#ifndef YY_DECL
#define YY_DECL int yyFlexLexer::yylex()
#endif

/* Code executed at the beginning of each rule, after yytext and yyleng
 * have been set up.
 */
#ifndef YY_USER_ACTION
#define YY_USER_ACTION
#endif

/* Code executed at the end of each rule. */
#ifndef YY_BREAK
#define YY_BREAK break;
#endif

#define YY_RULE_SETUP \
	YY_USER_ACTION

YY_DECL
	{
	register yy_state_type yy_current_state;
	register char *yy_cp, *yy_bp;
	register int yy_act;

#line 35 ".././ConfigLexer.l"


#line 835 "ConfigLexer.cpp"

	if ( yy_init )
		{
		yy_init = 0;

#ifdef YY_USER_INIT
		YY_USER_INIT;
#endif

		if ( ! yy_start )
			yy_start = 1;	/* first start state */

		if ( ! yyin )
			yyin = &std::cin;

		if ( ! yyout )
			yyout = &std::cout;

		if ( ! yy_current_buffer )
			yy_current_buffer =
				yy_create_buffer( yyin, YY_BUF_SIZE );

		yy_load_buffer_state();
		}

	while ( 1 )		/* loops until end-of-file is reached */
		{
		yy_cp = yy_c_buf_p;

		/* Support of yytext. */
		*yy_cp = yy_hold_char;

		/* yy_bp points to the position in yy_ch_buf of the start of
		 * the current run.
		 */
		yy_bp = yy_cp;

		yy_current_state = yy_start;
yy_match:
		do
			{
			register YY_CHAR yy_c = yy_ec[YY_SC_TO_UI(*yy_cp)];
			if ( yy_accept[yy_current_state] )
				{
				yy_last_accepting_state = yy_current_state;
				yy_last_accepting_cpos = yy_cp;
				}
			while ( yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state )
				{
				yy_current_state = (int) yy_def[yy_current_state];
				if ( yy_current_state >= 588 )
					yy_c = yy_meta[(unsigned int) yy_c];
				}
			yy_current_state = yy_nxt[yy_base[yy_current_state] + (unsigned int) yy_c];
			++yy_cp;
			}
		while ( yy_base[yy_current_state] != 668 );

yy_find_action:
		yy_act = yy_accept[yy_current_state];
		if ( yy_act == 0 )
			{ /* have to back up */
			yy_cp = yy_last_accepting_cpos;
			yy_current_state = yy_last_accepting_state;
			yy_act = yy_accept[yy_current_state];
			}

		YY_DO_BEFORE_ACTION;


do_action:	/* This label is used only to access EOF actions. */


		switch ( yy_act )
	{ /* beginning of action switch */
			case 0: /* must back up */
			/* undo the effects of YY_DO_BEFORE_ACTION */
			*yy_cp = yy_hold_char;
			yy_cp = yy_last_accepting_cpos;
			yy_current_state = yy_last_accepting_state;
			goto yy_find_action;

case 1:
YY_RULE_SETUP
#line 37 ".././ConfigLexer.l"
{
                        #ifdef DEBUG
                        char buff[128];
                        int i = 0;
                        #endif

                        char c;
                        while( (c = yyinput()) != '\n' )
                        {
                            if( c <= 0 )
                                break;
                            #ifdef DEBUG
                            buff[i++] = c;
                            #endif
                        }
                        #ifdef DEBUG
                        buff[i] = 0;
                        printf( "Single line comment: \"%s\"\n", buff );
                        #endif
                    }
	YY_BREAK
case 2:
YY_RULE_SETUP
#line 58 ".././ConfigLexer.l"
{ yylineno++;}
	YY_BREAK
case 3:
YY_RULE_SETUP
#line 59 ".././ConfigLexer.l"
{ ; }
	YY_BREAK
case 4:
YY_RULE_SETUP
#line 60 ".././ConfigLexer.l"
{ REPORT return '{'; }
	YY_BREAK
case 5:
YY_RULE_SETUP
#line 61 ".././ConfigLexer.l"
{ REPORT return '}'; }
	YY_BREAK
case 6:
YY_RULE_SETUP
#line 62 ".././ConfigLexer.l"
{
           		        char c;
	   		            int i = 0;
           		        while( (c = yyinput()) != '"' )
	       		            yytext[i++] = c;
	     		        yytext[i] = 0;
	    		        return PRTOKEN_QUOTED_STRING;
			        }
	YY_BREAK
case 7:
YY_RULE_SETUP
#line 70 ".././ConfigLexer.l"
{ REPORT return ','; }
	YY_BREAK
case 8:
YY_RULE_SETUP
#line 71 ".././ConfigLexer.l"
{ REPORT return ':'; }
	YY_BREAK
case 9:
YY_RULE_SETUP
#line 72 ".././ConfigLexer.l"
{ REPORT return ';'; }
	YY_BREAK
case 10:
YY_RULE_SETUP
#line 73 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_SET_SIMPLE; }
	YY_BREAK
case 11:
YY_RULE_SETUP
#line 74 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_VISUAL_ID; }
	YY_BREAK
case 12:
YY_RULE_SETUP
#line 75 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_BUFFER_SIZE; }
	YY_BREAK
case 13:
YY_RULE_SETUP
#line 76 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_LEVEL; }
	YY_BREAK
case 14:
YY_RULE_SETUP
#line 77 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_RGBA; }
	YY_BREAK
case 15:
YY_RULE_SETUP
#line 78 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_DOUBLEBUFFER; }
	YY_BREAK
case 16:
YY_RULE_SETUP
#line 79 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_STEREO; }
	YY_BREAK
case 17:
YY_RULE_SETUP
#line 80 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_AUX_BUFFERS; }
	YY_BREAK
case 18:
YY_RULE_SETUP
#line 81 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_RED_SIZE; }
	YY_BREAK
case 19:
YY_RULE_SETUP
#line 82 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_GREEN_SIZE; }
	YY_BREAK
case 20:
YY_RULE_SETUP
#line 83 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_BLUE_SIZE; }
	YY_BREAK
case 21:
YY_RULE_SETUP
#line 84 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_ALPHA_SIZE; }
	YY_BREAK
case 22:
YY_RULE_SETUP
#line 85 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_DEPTH_SIZE; }
	YY_BREAK
case 23:
YY_RULE_SETUP
#line 86 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_STENCIL_SIZE; }
	YY_BREAK
case 24:
YY_RULE_SETUP
#line 87 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_ACCUM_RED_SIZE; }
	YY_BREAK
case 25:
YY_RULE_SETUP
#line 88 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_ACCUM_GREEN_SIZE; }
	YY_BREAK
case 26:
YY_RULE_SETUP
#line 89 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_ACCUM_BLUE_SIZE; }
	YY_BREAK
case 27:
YY_RULE_SETUP
#line 90 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_ACCUM_ALPHA_SIZE; }
	YY_BREAK
case 28:
YY_RULE_SETUP
#line 91 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_SAMPLES; }
	YY_BREAK
case 29:
YY_RULE_SETUP
#line 92 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_SAMPLE_BUFFERS; }
	YY_BREAK
case 30:
YY_RULE_SETUP
#line 93 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_VISUAL; }
	YY_BREAK
case 31:
YY_RULE_SETUP
#line 94 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_RENDER_SURFACE; }
	YY_BREAK
case 32:
YY_RULE_SETUP
#line 95 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_WINDOW_RECT; }
	YY_BREAK
case 33:
YY_RULE_SETUP
#line 96 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_WINDOW_RECT; }
	YY_BREAK
case 34:
YY_RULE_SETUP
#line 97 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_HOSTNAME; }
	YY_BREAK
case 35:
YY_RULE_SETUP
#line 98 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_DISPLAY; }
	YY_BREAK
case 36:
YY_RULE_SETUP
#line 99 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_SCREEN; }
	YY_BREAK
case 37:
YY_RULE_SETUP
#line 100 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_BORDER; }
	YY_BREAK
case 38:
YY_RULE_SETUP
#line 101 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_INPUT_RECT; }
	YY_BREAK
case 39:
YY_RULE_SETUP
#line 102 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_INPUT_RECT; }
	YY_BREAK
case 40:
YY_RULE_SETUP
#line 103 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_DRAWABLE_TYPE; }
	YY_BREAK
case 41:
YY_RULE_SETUP
#line 104 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_PBUFFER_TYPE; }
	YY_BREAK
case 42:
YY_RULE_SETUP
#line 105 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_WINDOW_TYPE; }
	YY_BREAK
case 43:
YY_RULE_SETUP
#line 106 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_READ_DRAWABLE; }
	YY_BREAK
case 44:
YY_RULE_SETUP
#line 107 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_SET_RTT_MODE; }
	YY_BREAK
case 45:
YY_RULE_SETUP
#line 108 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_RTT_MODE_NONE; }
	YY_BREAK
case 46:
YY_RULE_SETUP
#line 109 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_RTT_MODE_RGB; }
	YY_BREAK
case 47:
YY_RULE_SETUP
#line 110 ".././ConfigLexer.l"
{ REPORT; return PRTOKEN_RTT_MODE_RGBA; }
	YY_BREAK
case 48:
YY_RULE_SETUP
#line 113 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_CAMERA_GROUP; }
	YY_BREAK
case 49:
YY_RULE_SETUP
#line 114 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_CAMERA; }
	YY_BREAK
case 50:
YY_RULE_SETUP
#line 115 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_PROJECTION_RECT; }
	YY_BREAK
case 51:
YY_RULE_SETUP
#line 116 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_PROJECTION_RECT; }
	YY_BREAK
case 52:
YY_RULE_SETUP
#line 118 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_LENS; }
	YY_BREAK
case 53:
YY_RULE_SETUP
#line 119 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_SHARELENS; }
	YY_BREAK
case 54:
YY_RULE_SETUP
#line 120 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_SHAREVIEW; }
	YY_BREAK
case 55:
YY_RULE_SETUP
#line 121 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_FRUSTUM; }
	YY_BREAK
case 56:
YY_RULE_SETUP
#line 122 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_PERSPECTIVE; }
	YY_BREAK
case 57:
YY_RULE_SETUP
#line 123 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_ORTHO; }
	YY_BREAK
case 58:
YY_RULE_SETUP
#line 124 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_CLEAR_COLOR; }
	YY_BREAK
case 59:
YY_RULE_SETUP
#line 126 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_OFFSET; }
	YY_BREAK
case 60:
YY_RULE_SETUP
#line 127 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_ROTATE; }
	YY_BREAK
case 61:
YY_RULE_SETUP
#line 128 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_TRANSLATE; }
	YY_BREAK
case 62:
YY_RULE_SETUP
#line 129 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_SCALE; }
	YY_BREAK
case 63:
YY_RULE_SETUP
#line 130 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_SHEAR; }
	YY_BREAK
case 64:
YY_RULE_SETUP
#line 131 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_METHOD; }
	YY_BREAK
case 65:
YY_RULE_SETUP
#line 132 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_PREMULTIPLY; }
	YY_BREAK
case 66:
YY_RULE_SETUP
#line 133 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_POSTMULTIPLY; }
	YY_BREAK
case 67:
YY_RULE_SETUP
#line 135 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_INPUT_AREA; }
	YY_BREAK
case 68:
YY_RULE_SETUP
#line 137 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_STEREO_SYSTEM_COMMANDS; }
	YY_BREAK
case 69:
YY_RULE_SETUP
#line 139 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_CUSTOM_FULL_SCREEN_RECTANGLE; }
	YY_BREAK
case 70:
YY_RULE_SETUP
#line 140 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_OVERRIDE_REDIRECT; }
	YY_BREAK
case 71:
YY_RULE_SETUP
#line 142 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_THREAD_MODEL; }
	YY_BREAK
case 72:
YY_RULE_SETUP
#line 143 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_SINGLE_THREADED; }
	YY_BREAK
case 73:
YY_RULE_SETUP
#line 144 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_THREAD_PER_CAMERA; }
	YY_BREAK
case 74:
YY_RULE_SETUP
#line 145 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_THREAD_PER_RENDER_SURFACE; }
	YY_BREAK
case 75:
YY_RULE_SETUP
#line 148 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_TRUE; }
	YY_BREAK
case 76:
YY_RULE_SETUP
#line 149 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_FALSE; }
	YY_BREAK
case 77:
YY_RULE_SETUP
#line 151 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_HEX_INTEGER; }
	YY_BREAK
case 78:
YY_RULE_SETUP
#line 152 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_INTEGER; }
	YY_BREAK
case 79:
YY_RULE_SETUP
#line 154 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_FLOAT; }
	YY_BREAK
case 80:
YY_RULE_SETUP
#line 155 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_FLOAT; }
	YY_BREAK
case 81:
YY_RULE_SETUP
#line 156 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_FLOAT; }
	YY_BREAK
case 82:
YY_RULE_SETUP
#line 157 ".././ConfigLexer.l"
{ REPORT return PRTOKEN_FLOAT; }
	YY_BREAK
case 83:
YY_RULE_SETUP
#line 162 ".././ConfigLexer.l"
ECHO;
	YY_BREAK
#line 1359 "ConfigLexer.cpp"
case YY_STATE_EOF(INITIAL):
	yyterminate();

	case YY_END_OF_BUFFER:
		{
		/* Amount of text matched not including the EOB char. */
		int yy_amount_of_matched_text = (int) (yy_cp - yytext_ptr) - 1;

		/* Undo the effects of YY_DO_BEFORE_ACTION. */
		*yy_cp = yy_hold_char;
		YY_RESTORE_YY_MORE_OFFSET

		if ( yy_current_buffer->yy_buffer_status == YY_BUFFER_NEW )
			{
			/* We're scanning a new file or input source.  It's
			 * possible that this happened because the user
			 * just pointed yyin at a new source and called
			 * yylex().  If so, then we have to assure
			 * consistency between yy_current_buffer and our
			 * globals.  Here is the right place to do so, because
			 * this is the first action (other than possibly a
			 * back-up) that will match for the new input source.
			 */
			yy_n_chars = yy_current_buffer->yy_n_chars;
			yy_current_buffer->yy_input_file = yyin;
			yy_current_buffer->yy_buffer_status = YY_BUFFER_NORMAL;
			}

		/* Note that here we test for yy_c_buf_p "<=" to the position
		 * of the first EOB in the buffer, since yy_c_buf_p will
		 * already have been incremented past the NUL character
		 * (since all states make transitions on EOB to the
		 * end-of-buffer state).  Contrast this with the test
		 * in input().
		 */
		if ( yy_c_buf_p <= &yy_current_buffer->yy_ch_buf[yy_n_chars] )
			{ /* This was really a NUL. */
			yy_state_type yy_next_state;

			yy_c_buf_p = yytext_ptr + yy_amount_of_matched_text;

			yy_current_state = yy_get_previous_state();

			/* Okay, we're now positioned to make the NUL
			 * transition.  We couldn't have
			 * yy_get_previous_state() go ahead and do it
			 * for us because it doesn't know how to deal
			 * with the possibility of jamming (and we don't
			 * want to build jamming into it because then it
			 * will run more slowly).
			 */

			yy_next_state = yy_try_NUL_trans( yy_current_state );

			yy_bp = yytext_ptr + YY_MORE_ADJ;

			if ( yy_next_state )
				{
				/* Consume the NUL. */
				yy_cp = ++yy_c_buf_p;
				yy_current_state = yy_next_state;
				goto yy_match;
				}

			else
				{
				yy_cp = yy_c_buf_p;
				goto yy_find_action;
				}
			}

		else switch ( yy_get_next_buffer() )
			{
			case EOB_ACT_END_OF_FILE:
				{
				yy_did_buffer_switch_on_eof = 0;

				if ( yywrap() )
					{
					/* Note: because we've taken care in
					 * yy_get_next_buffer() to have set up
					 * yytext, we can now set up
					 * yy_c_buf_p so that if some total
					 * hoser (like flex itself) wants to
					 * call the scanner after we return the
					 * YY_NULL, it'll still work - another
					 * YY_NULL will get returned.
					 */
					yy_c_buf_p = yytext_ptr + YY_MORE_ADJ;

					yy_act = YY_STATE_EOF(YY_START);
					goto do_action;
					}

				else
					{
					if ( ! yy_did_buffer_switch_on_eof )
						YY_NEW_FILE;
					}
				break;
				}

			case EOB_ACT_CONTINUE_SCAN:
				yy_c_buf_p =
					yytext_ptr + yy_amount_of_matched_text;

				yy_current_state = yy_get_previous_state();

				yy_cp = yy_c_buf_p;
				yy_bp = yytext_ptr + YY_MORE_ADJ;
				goto yy_match;

			case EOB_ACT_LAST_MATCH:
				yy_c_buf_p =
				&yy_current_buffer->yy_ch_buf[yy_n_chars];

				yy_current_state = yy_get_previous_state();

				yy_cp = yy_c_buf_p;
				yy_bp = yytext_ptr + YY_MORE_ADJ;
				goto yy_find_action;
			}
		break;
		}

	default:
		YY_FATAL_ERROR(
			"fatal flex scanner internal error--no action found" );
	} /* end of action switch */
		} /* end of scanning one token */
	} /* end of yylex */

yyFlexLexer::yyFlexLexer( std::istream* arg_yyin, std::ostream* arg_yyout )
	{
	yyin = arg_yyin;
	yyout = arg_yyout;
	yy_c_buf_p = 0;
	yy_init = 1;
	yy_start = 0;
	yy_flex_debug = 0;
	yylineno = 1;	// this will only get updated if %option yylineno

	yy_did_buffer_switch_on_eof = 0;

	yy_looking_for_trail_begin = 0;
	yy_more_flag = 0;
	yy_more_len = 0;
	yy_more_offset = yy_prev_more_offset = 0;

	yy_start_stack_ptr = yy_start_stack_depth = 0;
	yy_start_stack = 0;

	yy_current_buffer = 0;

#ifdef YY_USES_REJECT
	yy_state_buf = new yy_state_type[YY_BUF_SIZE + 2];
#else
	yy_state_buf = 0;
#endif

        yy_hold_char = 0;
        yy_n_chars = 0;
        yy_last_accepting_state = 0;
        yy_last_accepting_cpos = 0;

        yy_state_ptr = 0;
        yy_full_match = 0;
        yy_full_state = 0;
        yy_full_lp = 0;
	}

yyFlexLexer::~yyFlexLexer()
	{
	delete yy_state_buf;
	yy_delete_buffer( yy_current_buffer );
	}

void yyFlexLexer::switch_streams( std::istream* new_in, std::ostream* new_out )
	{
	if ( new_in )
		{
		yy_delete_buffer( yy_current_buffer );
		yy_switch_to_buffer( yy_create_buffer( new_in, YY_BUF_SIZE ) );
		}

	if ( new_out )
		yyout = new_out;
	}

#ifdef YY_INTERACTIVE
int yyFlexLexer::LexerInput( char* buf, int /* max_size */ )
#else
int yyFlexLexer::LexerInput( char* buf, int max_size )
#endif
	{
	if ( yyin->eof() || yyin->fail() )
		return 0;

#ifdef YY_INTERACTIVE
	yyin->get( buf[0] );

	if ( yyin->eof() )
		return 0;

	if ( yyin->bad() )
		return -1;

	return 1;

#else
	(void) yyin->read( buf, max_size );

	if ( yyin->bad() )
		return -1;
	else
		return yyin->gcount();
#endif
	}

void yyFlexLexer::LexerOutput( const char* buf, int size )
	{
	(void) yyout->write( buf, size );
	}

/* yy_get_next_buffer - try to read in a new buffer
 *
 * Returns a code representing an action:
 *	EOB_ACT_LAST_MATCH -
 *	EOB_ACT_CONTINUE_SCAN - continue scanning from current position
 *	EOB_ACT_END_OF_FILE - end of file
 */

int yyFlexLexer::yy_get_next_buffer()
	{
	register char *dest = yy_current_buffer->yy_ch_buf;
	register char *source = yytext_ptr;
	register int number_to_move, i;
	int ret_val;

	if ( yy_c_buf_p > &yy_current_buffer->yy_ch_buf[yy_n_chars + 1] )
		YY_FATAL_ERROR(
		"fatal flex scanner internal error--end of buffer missed" );

	if ( yy_current_buffer->yy_fill_buffer == 0 )
		{ /* Don't try to fill the buffer, so this is an EOF. */
		if ( yy_c_buf_p - yytext_ptr - YY_MORE_ADJ == 1 )
			{
			/* We matched a single character, the EOB, so
			 * treat this as a final EOF.
			 */
			return EOB_ACT_END_OF_FILE;
			}

		else
			{
			/* We matched some text prior to the EOB, first
			 * process it.
			 */
			return EOB_ACT_LAST_MATCH;
			}
		}

	/* Try to read more data. */

	/* First move last chars to start of buffer. */
	number_to_move = (int) (yy_c_buf_p - yytext_ptr) - 1;

	for ( i = 0; i < number_to_move; ++i )
		*(dest++) = *(source++);

	if ( yy_current_buffer->yy_buffer_status == YY_BUFFER_EOF_PENDING )
		/* don't do the read, it's not guaranteed to return an EOF,
		 * just force an EOF
		 */
		yy_current_buffer->yy_n_chars = yy_n_chars = 0;

	else
		{
		int num_to_read =
			yy_current_buffer->yy_buf_size - number_to_move - 1;

		while ( num_to_read <= 0 )
			{ /* Not enough room in the buffer - grow it. */
#ifdef YY_USES_REJECT
			YY_FATAL_ERROR(
"input buffer overflow, can't enlarge buffer because scanner uses REJECT" );
#else

			/* just a shorter name for the current buffer */
			YY_BUFFER_STATE b = yy_current_buffer;

			int yy_c_buf_p_offset =
				(int) (yy_c_buf_p - b->yy_ch_buf);

			if ( b->yy_is_our_buffer )
				{
				int new_size = b->yy_buf_size * 2;

				if ( new_size <= 0 )
					b->yy_buf_size += b->yy_buf_size / 8;
				else
					b->yy_buf_size *= 2;

				b->yy_ch_buf = (char *)
					/* Include room in for 2 EOB chars. */
					yy_flex_realloc( (void *) b->yy_ch_buf,
							 b->yy_buf_size + 2 );
				}
			else
				/* Can't grow it, we don't own it. */
				b->yy_ch_buf = 0;

			if ( ! b->yy_ch_buf )
				YY_FATAL_ERROR(
				"fatal error - scanner input buffer overflow" );

			yy_c_buf_p = &b->yy_ch_buf[yy_c_buf_p_offset];

			num_to_read = yy_current_buffer->yy_buf_size -
						number_to_move - 1;
#endif
			}

		if ( num_to_read > YY_READ_BUF_SIZE )
			num_to_read = YY_READ_BUF_SIZE;

		/* Read in more data. */
		YY_INPUT( (&yy_current_buffer->yy_ch_buf[number_to_move]),
			yy_n_chars, num_to_read );

		yy_current_buffer->yy_n_chars = yy_n_chars;
		}

	if ( yy_n_chars == 0 )
		{
		if ( number_to_move == YY_MORE_ADJ )
			{
			ret_val = EOB_ACT_END_OF_FILE;
			yyrestart( yyin );
			}

		else
			{
			ret_val = EOB_ACT_LAST_MATCH;
			yy_current_buffer->yy_buffer_status =
				YY_BUFFER_EOF_PENDING;
			}
		}

	else
		ret_val = EOB_ACT_CONTINUE_SCAN;

	yy_n_chars += number_to_move;
	yy_current_buffer->yy_ch_buf[yy_n_chars] = YY_END_OF_BUFFER_CHAR;
	yy_current_buffer->yy_ch_buf[yy_n_chars + 1] = YY_END_OF_BUFFER_CHAR;

	yytext_ptr = &yy_current_buffer->yy_ch_buf[0];

	return ret_val;
	}


/* yy_get_previous_state - get the state just before the EOB char was reached */

yy_state_type yyFlexLexer::yy_get_previous_state()
	{
	register yy_state_type yy_current_state;
	register char *yy_cp;

	yy_current_state = yy_start;

	for ( yy_cp = yytext_ptr + YY_MORE_ADJ; yy_cp < yy_c_buf_p; ++yy_cp )
		{
		register YY_CHAR yy_c = (*yy_cp ? yy_ec[YY_SC_TO_UI(*yy_cp)] : 1);
		if ( yy_accept[yy_current_state] )
			{
			yy_last_accepting_state = yy_current_state;
			yy_last_accepting_cpos = yy_cp;
			}
		while ( yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state )
			{
			yy_current_state = (int) yy_def[yy_current_state];
			if ( yy_current_state >= 588 )
				yy_c = yy_meta[(unsigned int) yy_c];
			}
		yy_current_state = yy_nxt[yy_base[yy_current_state] + (unsigned int) yy_c];
		}

	return yy_current_state;
	}


/* yy_try_NUL_trans - try to make a transition on the NUL character
 *
 * synopsis
 *	next_state = yy_try_NUL_trans( current_state );
 */

yy_state_type yyFlexLexer::yy_try_NUL_trans( yy_state_type yy_current_state )
	{
	register int yy_is_jam;
	register char *yy_cp = yy_c_buf_p;

	register YY_CHAR yy_c = 1;
	if ( yy_accept[yy_current_state] )
		{
		yy_last_accepting_state = yy_current_state;
		yy_last_accepting_cpos = yy_cp;
		}
	while ( yy_chk[yy_base[yy_current_state] + yy_c] != yy_current_state )
		{
		yy_current_state = (int) yy_def[yy_current_state];
		if ( yy_current_state >= 588 )
			yy_c = yy_meta[(unsigned int) yy_c];
		}
	yy_current_state = yy_nxt[yy_base[yy_current_state] + (unsigned int) yy_c];
	yy_is_jam = (yy_current_state == 587);

	return yy_is_jam ? 0 : yy_current_state;
	}


void yyFlexLexer::yyunput( int c, register char* yy_bp )
	{
	register char *yy_cp = yy_c_buf_p;

	/* undo effects of setting up yytext */
	*yy_cp = yy_hold_char;

	if ( yy_cp < yy_current_buffer->yy_ch_buf + 2 )
		{ /* need to shift things up to make room */
		/* +2 for EOB chars. */
		register int number_to_move = yy_n_chars + 2;
		register char *dest = &yy_current_buffer->yy_ch_buf[
					yy_current_buffer->yy_buf_size + 2];
		register char *source =
				&yy_current_buffer->yy_ch_buf[number_to_move];

		while ( source > yy_current_buffer->yy_ch_buf )
			*--dest = *--source;

		yy_cp += (int) (dest - source);
		yy_bp += (int) (dest - source);
		yy_current_buffer->yy_n_chars =
			yy_n_chars = yy_current_buffer->yy_buf_size;

		if ( yy_cp < yy_current_buffer->yy_ch_buf + 2 )
			YY_FATAL_ERROR( "flex scanner push-back overflow" );
		}

	*--yy_cp = (char) c;


	yytext_ptr = yy_bp;
	yy_hold_char = *yy_cp;
	yy_c_buf_p = yy_cp;
	}


int yyFlexLexer::yyinput()
	{
	int c;

	*yy_c_buf_p = yy_hold_char;

	if ( *yy_c_buf_p == YY_END_OF_BUFFER_CHAR )
		{
		/* yy_c_buf_p now points to the character we want to return.
		 * If this occurs *before* the EOB characters, then it's a
		 * valid NUL; if not, then we've hit the end of the buffer.
		 */
		if ( yy_c_buf_p < &yy_current_buffer->yy_ch_buf[yy_n_chars] )
			/* This was really a NUL. */
			*yy_c_buf_p = '\0';

		else
			{ /* need more input */
			int offset = yy_c_buf_p - yytext_ptr;
			++yy_c_buf_p;

			switch ( yy_get_next_buffer() )
				{
				case EOB_ACT_LAST_MATCH:
					/* This happens because yy_g_n_b()
					 * sees that we've accumulated a
					 * token and flags that we need to
					 * try matching the token before
					 * proceeding.  But for input(),
					 * there's no matching to consider.
					 * So convert the EOB_ACT_LAST_MATCH
					 * to EOB_ACT_END_OF_FILE.
					 */

					/* Reset buffer status. */
					yyrestart( yyin );

					/* fall through */

				case EOB_ACT_END_OF_FILE:
					{
					if ( yywrap() )
						return EOF;

					if ( ! yy_did_buffer_switch_on_eof )
						YY_NEW_FILE;
#ifdef __cplusplus
					return yyinput();
#else
					return input();
#endif
					}

				case EOB_ACT_CONTINUE_SCAN:
					yy_c_buf_p = yytext_ptr + offset;
					break;
				}
			}
		}

	c = *(unsigned char *) yy_c_buf_p;	/* cast for 8-bit char's */
	*yy_c_buf_p = '\0';	/* preserve yytext */
	yy_hold_char = *++yy_c_buf_p;


	return c;
	}


void yyFlexLexer::yyrestart( std::istream* input_file )
	{
	if ( ! yy_current_buffer )
		yy_current_buffer = yy_create_buffer( yyin, YY_BUF_SIZE );

	yy_init_buffer( yy_current_buffer, input_file );
	yy_load_buffer_state();
	}


void yyFlexLexer::yy_switch_to_buffer( YY_BUFFER_STATE new_buffer )
	{
	if ( yy_current_buffer == new_buffer )
		return;

	if ( yy_current_buffer )
		{
		/* Flush out information for old buffer. */
		*yy_c_buf_p = yy_hold_char;
		yy_current_buffer->yy_buf_pos = yy_c_buf_p;
		yy_current_buffer->yy_n_chars = yy_n_chars;
		}

	yy_current_buffer = new_buffer;
	yy_load_buffer_state();

	/* We don't actually know whether we did this switch during
	 * EOF (yywrap()) processing, but the only time this flag
	 * is looked at is after yywrap() is called, so it's safe
	 * to go ahead and always set it.
	 */
	yy_did_buffer_switch_on_eof = 1;
	}


void yyFlexLexer::yy_load_buffer_state()
	{
	yy_n_chars = yy_current_buffer->yy_n_chars;
	yytext_ptr = yy_c_buf_p = yy_current_buffer->yy_buf_pos;
	yyin = yy_current_buffer->yy_input_file;
	yy_hold_char = *yy_c_buf_p;
	}


YY_BUFFER_STATE yyFlexLexer::yy_create_buffer( std::istream* file, int size )
	{
	YY_BUFFER_STATE b;

	b = (YY_BUFFER_STATE) yy_flex_alloc( sizeof( struct yy_buffer_state ) );
	if ( ! b )
        {
		YY_FATAL_ERROR( "out of dynamic memory in yy_create_buffer()" );
		return 0;
        }

	b->yy_buf_size = size;

	/* yy_ch_buf has to be 2 characters longer than the size given because
	 * we need to put in 2 end-of-buffer characters.
	 */
	b->yy_ch_buf = (char *) yy_flex_alloc( b->yy_buf_size + 2 );
	if ( ! b->yy_ch_buf )
		YY_FATAL_ERROR( "out of dynamic memory in yy_create_buffer()" );

	b->yy_is_our_buffer = 1;

	yy_init_buffer( b, file );

	return b;
	}


void yyFlexLexer::yy_delete_buffer( YY_BUFFER_STATE b )
	{
	if ( ! b )
		return;

	if ( b == yy_current_buffer )
		yy_current_buffer = (YY_BUFFER_STATE) 0;

	if ( b->yy_is_our_buffer )
		yy_flex_free( (void *) b->yy_ch_buf );

	yy_flex_free( (void *) b );
	}

#ifndef __THROW
    #define __THROW
#endif

extern "C" int isatty YY_PROTO(( int )) __THROW;

void yyFlexLexer::yy_init_buffer( YY_BUFFER_STATE b, std::istream* file )

	{
	yy_flush_buffer( b );

	b->yy_input_file = file;
	b->yy_fill_buffer = 1;

	b->yy_is_interactive = 0;
	}


void yyFlexLexer::yy_flush_buffer( YY_BUFFER_STATE b )
	{
	if ( ! b )
		return;

	b->yy_n_chars = 0;

	/* We always need two end-of-buffer characters.  The first causes
	 * a transition to the end-of-buffer state.  The second causes
	 * a jam in that state.
	 */
	b->yy_ch_buf[0] = YY_END_OF_BUFFER_CHAR;
	b->yy_ch_buf[1] = YY_END_OF_BUFFER_CHAR;

	b->yy_buf_pos = &b->yy_ch_buf[0];

	b->yy_at_bol = 1;
	b->yy_buffer_status = YY_BUFFER_NEW;

	if ( b == yy_current_buffer )
		yy_load_buffer_state();
	}


#ifndef YY_NO_SCAN_BUFFER
#endif


#ifndef YY_NO_SCAN_STRING
#endif


#ifndef YY_NO_SCAN_BYTES
#endif


#ifndef YY_NO_PUSH_STATE
void yyFlexLexer::yy_push_state( int new_state )
	{
	if ( yy_start_stack_ptr >= yy_start_stack_depth )
		{
		yy_size_t new_size;

		yy_start_stack_depth += YY_START_STACK_INCR;
		new_size = yy_start_stack_depth * sizeof( int );

		if ( ! yy_start_stack )
			yy_start_stack = (int *) yy_flex_alloc( new_size );

		else
			yy_start_stack = (int *) yy_flex_realloc(
					(void *) yy_start_stack, new_size );

		if ( ! yy_start_stack )
			YY_FATAL_ERROR(
			"out of memory expanding start-condition stack" );
		}

	yy_start_stack[yy_start_stack_ptr++] = YY_START;

	BEGIN(new_state);
	}
#endif


#ifndef YY_NO_POP_STATE
void yyFlexLexer::yy_pop_state()
	{
	if ( --yy_start_stack_ptr < 0 )
		YY_FATAL_ERROR( "start-condition stack underflow" );

	BEGIN(yy_start_stack[yy_start_stack_ptr]);
	}
#endif


#ifndef YY_NO_TOP_STATE
int yyFlexLexer::yy_top_state()
	{
	return yy_start_stack[yy_start_stack_ptr - 1];
	}
#endif

#ifndef YY_EXIT_FAILURE
#define YY_EXIT_FAILURE 2
#endif


void yyFlexLexer::LexerError( yyconst char msg[] )
	{
	std::cerr << msg << '\n';
	exit( YY_EXIT_FAILURE );
	}


/* Redefine yyless() so it works in section 3 code. */

#undef yyless
#define yyless(n) \
	do \
		{ \
		/* Undo effects of setting up yytext. */ \
		yytext[yyleng] = yy_hold_char; \
		yy_c_buf_p = yytext + n; \
		yy_hold_char = *yy_c_buf_p; \
		*yy_c_buf_p = '\0'; \
		yyleng = n; \
		} \
	while ( 0 )


/* Internal utility routines. */

#ifndef yytext_ptr
#ifdef YY_USE_PROTOS
static void yy_flex_strncpy( char *s1, yyconst char *s2, int n )
#else
static void yy_flex_strncpy( s1, s2, n )
char *s1;
yyconst char *s2;
int n;
#endif
	{
	register int i;
	for ( i = 0; i < n; ++i )
		s1[i] = s2[i];
	}
#endif

#ifdef YY_NEED_STRLEN
#ifdef YY_USE_PROTOS
static int yy_flex_strlen( yyconst char *s )
#else
static int yy_flex_strlen( s )
yyconst char *s;
#endif
	{
	register int n;
	for ( n = 0; s[n]; ++n )
		;

	return n;
	}
#endif


#ifdef YY_USE_PROTOS
static void *yy_flex_alloc( yy_size_t size )
#else
static void *yy_flex_alloc( size )
yy_size_t size;
#endif
	{
	return (void *) malloc( size );
	}

#ifdef YY_USE_PROTOS
static void *yy_flex_realloc( void *ptr, yy_size_t size )
#else
static void *yy_flex_realloc( ptr, size )
void *ptr;
yy_size_t size;
#endif
	{
	/* The cast to (char *) in the following accommodates both
	 * implementations that use char* generic pointers, and those
	 * that use void* generic pointers.  It works with the latter
	 * because both ANSI C and C++ allow castless assignment from
	 * any pointer type to void*, and deal with argument conversions
	 * as though doing an assignment.
	 */
	return (void *) realloc( (char *) ptr, size );
	}

#ifdef YY_USE_PROTOS
static void yy_flex_free( void *ptr )
#else
static void yy_flex_free( ptr )
void *ptr;
#endif
	{
	free( ptr );
	}

#if YY_MAIN
int main()
	{
	yylex();
	return 0;
	}
#endif
#line 162 ".././ConfigLexer.l"



