#include <osgDB/Registry>

USE_SERIALIZER_WRAPPER(osgSim_AzimElevationSector)
USE_SERIALIZER_WRAPPER(osgSim_AzimSector)
USE_SERIALIZER_WRAPPER(osgSim_BlinkSequence)
USE_SERIALIZER_WRAPPER(osgSim_ConeSector)
USE_SERIALIZER_WRAPPER(osgSim_DirectionalSector)
USE_SERIALIZER_WRAPPER(osgSim_DOFTransform)
USE_SERIALIZER_WRAPPER(osgSim_ElevationSector)
USE_SERIALIZER_WRAPPER(osgSim_Impostor)
USE_SERIALIZER_WRAPPER(osgSim_LightPointNode)
USE_SERIALIZER_WRAPPER(osgSim_LightPointSystem)
USE_SERIALIZER_WRAPPER(osgSim_MultiSwitch)
USE_SERIALIZER_WRAPPER(osgSim_ObjectRecordData)
USE_SERIALIZER_WRAPPER(osgSim_OverlayNode)
USE_SERIALIZER_WRAPPER(osgSim_ScalarBar)
USE_SERIALIZER_WRAPPER(osgSim_Sector)
USE_SERIALIZER_WRAPPER(osgSim_SequenceGroup)
USE_SERIALIZER_WRAPPER(osgSim_ShapeAttributeList)
USE_SERIALIZER_WRAPPER(osgSim_SphereSegment)
USE_SERIALIZER_WRAPPER(osgSim_VisibilityGroup)

extern "C" void wrapper_serializer_library_osgSim(void) {}

