#ifndef ESRITYPE_H_
#define ESRITYPE_H_


namespace ESRIShape {

typedef int             Integer;
typedef short           Short;
typedef unsigned char   Byte;
typedef double          Double;
typedef unsigned char * BytePtr;

}

#endif /*ESRITYPE_H_*/
