/* -*-c++-*- OpenSceneGraph - Copyright (C) 1998-2006 Robert Osfield
 *
 * This library is open source and may be redistributed and/or modified under
 * the terms of the OpenSceneGraph Public License (OSGPL) version 0.0 or
 * (at your option) any later version.  The full license is in LICENSE file
 * included with this distribution, and on the openscenegraph.org website.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * OpenSceneGraph Public License for more details.
*/

//
// OpenFlightŪ loader for OpenSceneGraph
//
//  Copyright (C) 2005-2007  Brede Johansen
//

#include "Registry.h"

using namespace flt;

// Prevent "unknown record" message for the following reserved records:
REGISTER_FLTRECORD(DummyRecord, 103)
REGISTER_FLTRECORD(DummyRecord, 104)
REGISTER_FLTRECORD(DummyRecord, 117)
REGISTER_FLTRECORD(DummyRecord, 118)
REGISTER_FLTRECORD(DummyRecord, 120)
REGISTER_FLTRECORD(DummyRecord, 121)
REGISTER_FLTRECORD(DummyRecord, 124)
REGISTER_FLTRECORD(DummyRecord, 125)





