#include <osgDB/Registry>

USE_SERIALIZER_WRAPPER(osgParticleAccelOperator)
USE_SERIALIZER_WRAPPER(osgParticleAngularAccelOperator)
USE_SERIALIZER_WRAPPER(osgParticleAngularDampingOperator)
USE_SERIALIZER_WRAPPER(osgParticleBounceOperator)
USE_SERIALIZER_WRAPPER(osgParticleBoxPlacer)
USE_SERIALIZER_WRAPPER(osgParticleCenteredPlacer)
USE_SERIALIZER_WRAPPER(osgParticleCompositePlacer)
USE_SERIALIZER_WRAPPER(osgParticleConnectedParticleSystem)
USE_SERIALIZER_WRAPPER(osgParticleConstantRateCounter)
USE_SERIALIZER_WRAPPER(osgParticleCounter)
USE_SERIALIZER_WRAPPER(osgParticleDampingOperator)
USE_SERIALIZER_WRAPPER(osgParticleDomainOperator)
USE_SERIALIZER_WRAPPER(osgParticleEmitter)
USE_SERIALIZER_WRAPPER(osgParticleExplosionDebrisEffect)
USE_SERIALIZER_WRAPPER(osgParticleExplosionEffect)
USE_SERIALIZER_WRAPPER(osgParticleExplosionOperator)
USE_SERIALIZER_WRAPPER(osgParticleFireEffect)
USE_SERIALIZER_WRAPPER(osgParticleFluidFrictionOperator)
USE_SERIALIZER_WRAPPER(osgParticleFluidProgram)
USE_SERIALIZER_WRAPPER(osgParticleForceOperator)
USE_SERIALIZER_WRAPPER(osgParticleInterpolator)
USE_SERIALIZER_WRAPPER(osgParticleLinearInterpolator)
USE_SERIALIZER_WRAPPER(osgParticleModularEmitter)
USE_SERIALIZER_WRAPPER(osgParticleModularProgram)
USE_SERIALIZER_WRAPPER(osgParticleMultiSegmentPlacer)
USE_SERIALIZER_WRAPPER(osgParticleOperator)
USE_SERIALIZER_WRAPPER(osgParticleOrbitOperator)
USE_SERIALIZER_WRAPPER(osgParticleParticleEffect)
USE_SERIALIZER_WRAPPER(osgParticleParticleProcessor)
USE_SERIALIZER_WRAPPER(osgParticleParticleSystem)
USE_SERIALIZER_WRAPPER(osgParticleParticleSystemUpdater)
USE_SERIALIZER_WRAPPER(osgParticlePlacer)
USE_SERIALIZER_WRAPPER(osgParticlePointPlacer)
USE_SERIALIZER_WRAPPER(osgParticlePrecipitationEffect)
USE_SERIALIZER_WRAPPER(osgParticleProgram)
USE_SERIALIZER_WRAPPER(osgParticleRadialShooter)
USE_SERIALIZER_WRAPPER(osgParticleRandomRateCounter)
USE_SERIALIZER_WRAPPER(osgParticleSectorPlacer)
USE_SERIALIZER_WRAPPER(osgParticleSegmentPlacer)
USE_SERIALIZER_WRAPPER(osgParticleShooter)
USE_SERIALIZER_WRAPPER(osgParticleSinkOperator)
USE_SERIALIZER_WRAPPER(osgParticleSmokeEffect)
USE_SERIALIZER_WRAPPER(osgParticleSmokeTrailEffect)
USE_SERIALIZER_WRAPPER(osgParticleVariableRateCounter)

extern "C" void wrapper_serializer_library_osgParticle(void) {}

