char osgText_Text_vert[] = "$OSG_GLSL_VERSION\n"
                           "$OSG_PRECISION_FLOAT\n"
                           "\n"
                           "$OSG_VARYING_OUT vec2 texCoord;\n"
                           "$OSG_VARYING_OUT vec4 vertexColor;\n"
                           "\n"
                           "void main(void)\n"
                           "{\n"
                           "    gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;\n"
                           "    texCoord = gl_MultiTexCoord0.xy;\n"
                           "    vertexColor = gl_Color;\n"
                           "\n"
                           "#if !defined(GL_ES) && __VERSION__<140\n"
                           "    gl_ClipVertex = gl_ModelViewMatrix * gl_Vertex;\n"
                           "#endif\n"
                           "}\n"
                           "\n";
