#include <osg/ShaderAttribute>
#include <osgDB/ObjectWrapper>
#include <osgDB/InputStream>
#include <osgDB/OutputStream>

REGISTER_OBJECT_WRAPPER( ShaderAttribute,
                         /*new osg::ShaderAttribute*/NULL,
                         osg::ShaderAttribute,
                         "osg::Object osg::StateAttribute osg::ShaderAttribute" )
{
}
