#include <osgDB/Registry>

USE_DOTOSGWRAPPER(g_osgWidget_BoxProxy)
USE_DOTOSGWRAPPER(g_osgWidget_EmbeddedWindowProxy)
USE_DOTOSGWRAPPER(g_osgWidget_FrameProxy)
USE_DOTOSGWRAPPER(g_osgWidget_InputProxy)
USE_DOTOSGWRAPPER(g_osgWidget_LabelProxy)
USE_DOTOSGWRAPPER(g_osgWidget_TableProxy)
USE_DOTOSGWRAPPER(g_osgWidget_WidgetProxy)
USE_DOTOSGWRAPPER(g_osgWidget_NotifyWidgetProxy)
USE_DOTOSGWRAPPER(g_osgWidget_NullWidgetProxy)
USE_DOTOSGWRAPPER(g_osgWidget_WindowManagerProxy)

extern "C" void dotosgwrapper_library_osgWidget(void) {}

