char shadergen_frag[] = "// ShaderGen shader\n"
                        "// new version\n"
                        "#ifdef GL_ES\n"
                        "    precision highp float;\n"
                        "#endif\n"
                        "\n"
                        "#pragma import_defines(GL_TEXTURE_2D)\n"
                        "\n"
                        "varying vec4 vertexColor;\n"
                        "\n"
                        "#if defined(GL_TEXTURE_2D)\n"
                        "uniform sampler2D diffuseMap;\n"
                        "#endif\n"
                        "\n"
                        "void main()\n"
                        "{\n"
                        "  vec4 color = vertexColor;\n"
                        "\n"
                        "#if defined(GL_TEXTURE_2D)\n"
                        "  color = color * texture2D(diffuseMap, gl_TexCoord[0].st);\n"
                        "#endif\n"
                        "\n"
                        "  gl_FragColor = color;\n"
                        "}\n"
                        "\n";
