#include <osgDB/Registry>

USE_DOTOSGWRAPPER(AccelOperator_Proxy)
USE_DOTOSGWRAPPER(AngularAccelOperator_Proxy)
USE_DOTOSGWRAPPER(AngularDampingOperator_Proxy)
USE_DOTOSGWRAPPER(BounceOperator_Proxy)
USE_DOTOSGWRAPPER(BoxPlacer_Proxy)
USE_DOTOSGWRAPPER(CenteredPlacer_Proxy)
USE_DOTOSGWRAPPER(ConnectedParticleSystem_Proxy)
USE_DOTOSGWRAPPER(ConstantRateCounter_Proxy)
USE_DOTOSGWRAPPER(DampingOperator_Proxy)
USE_DOTOSGWRAPPER(DomainOperator_Proxy)
USE_DOTOSGWRAPPER(Emitter_Proxy)
USE_DOTOSGWRAPPER(ExplosionDebrisEffect_Proxy)
USE_DOTOSGWRAPPER(ExplosionEffect_Proxy)
USE_DOTOSGWRAPPER(ExplosionOperator_Proxy)
USE_DOTOSGWRAPPER(FireEffect_Proxy)
USE_DOTOSGWRAPPER(FluidFrictionOperator_Proxy)
USE_DOTOSGWRAPPER(FluidProgram_Proxy)
USE_DOTOSGWRAPPER(ForceOperator_Proxy)
USE_DOTOSGWRAPPER(LinearInterpolator_Proxy)
USE_DOTOSGWRAPPER(ModularEmitter_Proxy)
USE_DOTOSGWRAPPER(ModularProgram_Proxy)
USE_DOTOSGWRAPPER(MultiSegmentPlacer_Proxy)
USE_DOTOSGWRAPPER(OrbitOperator_Proxy)
USE_DOTOSGWRAPPER(ParticleEffect_Proxy)
USE_DOTOSGWRAPPER(ParticleProcessor_Proxy)
USE_DOTOSGWRAPPER(ParticleSystem_Proxy)
USE_DOTOSGWRAPPER(PSU_Proxy)
USE_DOTOSGWRAPPER(PointPlacer_Proxy)
USE_DOTOSGWRAPPER(g_IOProgramProxy)
USE_DOTOSGWRAPPER(RadialShooter_Proxy)
USE_DOTOSGWRAPPER(RandomRateCounter_Proxy)
USE_DOTOSGWRAPPER(SectorPlacer_Proxy)
USE_DOTOSGWRAPPER(SegmentPlacer_Proxy)
USE_DOTOSGWRAPPER(SinkOperator_Proxy)
USE_DOTOSGWRAPPER(SmokeEffect_Proxy)
USE_DOTOSGWRAPPER(SmokeTrailEffect_Proxy)
USE_DOTOSGWRAPPER(VariableRateCounter_Proxy)

extern "C" void dotosgwrapper_library_osgParticle(void) {}

